UNIT BinIO;

INTERFACE

Uses SysUtils, Classes, ZLib;

Const BufMax = 64*1024;

Type TBinRead = class
     private
        f: TFileStream;          // vstup ze souboru
        c: TDecompressionStream; // dekompresni proud

        buffered:   boolean;     // true => pouziva se vyrovnavaci pameti
        compressed: boolean;     // true => pouziva se komprese

        comp_len:   integer;     // pocet zbyvajicich nekoprimovanych bytu na vstupu

        pos: integer;            // pocet prectenych bytu z vyrovnavaci pameti
        len: integer;            // pocet bytu ve vyrovnavaci pameti

        buf: array [0..BufMax-1] of byte; // vyrovnavaci pamet

        Procedure ReadBuf;
        Procedure ReadData (adr: pointer; size: integer);

     public
        Procedure Open (FileName: string);
        Procedure Close;

        Procedure StartBuffering;
        Procedure StartDecompression;

        Procedure ReadBool (var x: boolean);
        Procedure ReadByte (var x: byte);

        Procedure ReadInt  (var x: integer);
        Procedure ReadOrd  (var x: longword);
        Procedure ReadCode (var x: integer);

        Procedure ReadStr (var s: string);
        Procedure ReadBlock (var adr: pointer; var size: integer);
     end;

Type TBinWrite = class
     private
        f: TFileStream;        // vystup do souboru
        c: TCompressionStream; // kompresni proud

        buffered:   boolean;   // true => pouziva se vyrovnavaci pameti
        compressed: boolean;   // true => pouziva se komprese

        comp_ofs: integer;     // pozice v souboru, kde bude ulozeno comp_len
        comp_len: integer;     // pocet bytu pred kompresi

        len: integer;          // pocet bytu ve vyrovnavaci pameti
        buf: array [0..BufMax-1] of byte; // vyrovnavaci pamet

        Procedure WriteBuf;
        Procedure WriteData (adr: pointer; size: integer);

     public
        Procedure Open (FileName: string);
        Procedure Close;

        Procedure StartBuffering;
        Procedure StartCompression;

        Procedure WriteBool (x: boolean);
        Procedure WriteByte (x: byte);

        Procedure WriteInt  (x: integer);
        Procedure WriteOrd  (x: longword);
        Procedure WriteCode (x: integer);

        Procedure WriteStr (s: string);
        Procedure WriteBlock (adr: pointer; size: integer);
     end;

IMPLEMENTATION

(********************************* ERROR **********************************)

Procedure Error (s: string);
Begin
     raise Exception.Create (s);
End;

(****************************** OPEN / CLOSE ******************************)

Procedure TBinRead.Open (FileName: string);
Begin
     buffered := false;
     compressed := false;

     comp_len := 0;

     pos := 0;
     len := 0;

     f := TFileStream.Create (FileName, fmOpenRead);
     c := nil;
End;

Procedure TBinRead.Close;
Begin
     if compressed then begin
        c.Free;
        c := nil;
     end;

     f.Free;
     f := nil;
End;

{--------------------------------------------------------------------------}

Procedure TBinWrite.Open (FileName: string);
Begin
     buffered := false;
     compressed := false;

     comp_ofs := 0;
     comp_len := 0;

     len := 0;

     f := TFileStream.Create (FileName, fmCreate);
End;

Procedure TBinWrite.Close;
Begin
     if len > 0 then WriteBuf;

     if compressed then begin
        c.Free;
        c := nil;

        // navrat na zacatek a zapis delky
        f.Seek (comp_ofs, soFromBeginning);
        f.WriteBuffer (comp_len, sizeof (comp_len));
     end;

     f.Free;
     f := nil;
End;

(************************ COMPRESSION / BUFFERING *************************)

{ na zacatku souboru se cte/zapisuje bez vyrovnavaci pameti (buffered=false, compressed=false)
  potom lze prepnout do rezimu s vyrovnavaci pameti (buffered=true, compressed=false)
  nebo do rezimu s kompresi a vyrovnavaci pameti (buffered=true, compressed=true) }

Procedure TBinRead.StartBuffering;
Begin
     if buffered or compressed then Error ('Invalid mode');
     buffered := true;
End;

Procedure TBinRead.StartDecompression;
Begin
     if buffered or compressed then Error ('Invalid mode');

     // precti puvodni pocet bytu
     f.ReadBuffer (comp_len, sizeof (comp_len));

     buffered := true;
     compressed := true;
     c := TDecompressionStream.Create (f);
End;

{--------------------------------------------------------------------------}

Procedure TBinWrite.StartBuffering;
Begin
     if buffered or compressed then Error ('Invalid mode');
     buffered := true;
End;

Procedure TBinWrite.StartCompression;
Begin
     if buffered or compressed then Error ('Invalid mode');

     // misto pro pocet bytu pred kompresi
     comp_ofs := f.Position;
     comp_len := 0;
     f.WriteBuffer (comp_len, sizeof (comp_len));

     buffered := true;
     compressed := true;
     c := TCompressionStream.Create (clDefault, f);
End;

(****************************** READ / WRITE ******************************)

Procedure TBinRead.ReadBuf;
Var  lim: integer;
Begin
     if not buffered then Error ('Internal error');

     if compressed then begin
        lim := BufMax;
        if lim > comp_len then lim := comp_len;
        if lim = 0 then Error ('End of file');
        len := c.Read (buf, lim);
        dec (comp_len, lim);
     end
     else begin
        len := f.Read (buf, BufMax);
     end;

     pos := 0;
End;

Procedure TBinRead.ReadData (adr: pointer; size: integer);
Var  step: integer;
Begin
     if not buffered then
        f.ReadBuffer (adr^, size)

     else begin
        while size > 0 do begin
           if pos = len then begin
              ReadBuf;
              if len = 0 then Error ('End of file');
           end;

           step := len - pos;
           if step > size then step := size;

           move (buf [pos], adr^, step);
           inc (pos, step);
           dec (size, step);
           inc (integer (adr), step);

           if (step <= 0) or (size < 0) or
              (pos < 0) or (pos > len) or
              (len < 0) or (len > BufMax) then
              Error ('Internal error');
        end;
     end

End;

{--------------------------------------------------------------------------}

Procedure TBinWrite.WriteBuf;
Begin
     if not buffered then Error ('Internal error');

     if compressed then begin
        c.WriteBuffer (buf, len);
        inc (comp_len, len);
     end
     else begin
        f.WriteBuffer (buf, len);
     end;

     len := 0;
End;

Procedure TBinWrite.WriteData (adr: pointer; size: integer);
Var  step: integer;
Begin
     if not buffered then
        f.WriteBuffer (adr^, size)

     else begin
        while size > 0 do begin
           step := BufMax - len;
           if step > size then step := size;

           move (adr^, buf [len], step);
           inc (len, step);
           dec (size, step);
           inc (integer (adr), step);

           if (step <= 0) or (size < 0) or
              (len < 0) or (len > BufMax) then
              Error ('Internal error');

           if len = BufMax then WriteBuf;
        end;
     end;
End;

(***************************** READ VARIABLE ******************************)

Procedure TBinRead.ReadBool (var x: boolean);
Begin
     ReadData (@x, sizeof (x));
End;

Procedure TBinRead.ReadByte (var x: byte);
Begin
     ReadData (@x, sizeof (x));
End;

Procedure TBinRead.ReadInt (var x: integer);
Begin
     ReadData (@x, sizeof (x));
End;

Procedure TBinRead.ReadOrd (var x: longword);
Begin
     ReadData (@x, sizeof (x));
End;

Procedure TBinRead.ReadCode (var x: integer);
Var  n: byte;
Begin
     ReadData (@n, 1);
     if n < 255 then x := n
                else ReadData (@x, sizeof (x));
End;

Procedure TBinRead.ReadStr (var s: string);
Var  n: integer;
Begin
     ReadCode (n);
     SetString (s, nil, n);
     ReadData (PChar (s), n);
End;

Procedure TBinRead.ReadBlock (var adr: pointer; var size: integer);
Begin
     ReadCode (size);
     adr := AllocMem (size); { allocate memory }
     ReadData (adr, size);
End;

(***************************** WRITE VARIABLE *****************************)

Procedure TBinWrite.WriteBool (x: boolean);
Begin
     WriteData (@x, sizeof (x));
End;

Procedure TBinWrite.WriteByte (x: byte);
Begin
     WriteData (@x, sizeof (x));
End;

Procedure TBinWrite.WriteInt (x: integer);
Begin
     WriteData (@x, sizeof (x));
End;

Procedure TBinWrite.WriteOrd (x: longword);
Begin
     WriteData (@x, sizeof (x));
End;

Procedure TBinWrite.WriteCode (x: integer);
Var  n: byte;
Begin
     if (x >= 0) and (x <= 254) then
        WriteData (@x, 1)
     else begin
        n := 255;
        WriteData (@n, 1);
        WriteData (@x, sizeof (x));
     end;
End;

Procedure TBinWrite.WriteStr (s: string);
Var  n: integer;
Begin
     n := length (s);
     WriteCode (n);
     WriteData (PChar (s), n);
End;

Procedure TBinWrite.WriteBlock (adr: pointer; size: integer);
Begin
     WriteCode (size);
     WriteData (adr, size);
End;

END.

