UNIT TxtIO;

INTERFACE

Uses SysUtils;

Const cr = ^M;
      lf = ^J;

Var InpEof:    boolean;      { true => end of file }
    InpLine:   string;       { input line }
    InpPos:    integer;      { position of current character }
    InpChar:   char;         { current character }

Procedure GetChar;
Function  PeekChar (i: integer): char;
Procedure GetLine;

Procedure OpenInp (FileName: string);
Procedure CloseInp;
Procedure EraseInp;

Procedure Space;

Function GetName: string;
Function GetCommaName: string;
Function GetInt: integer;
Function GetHex: integer;

Function GetAttr: integer;
Function GetStdDateTime: integer;
Function GetFlexDateTime: integer;

Procedure SkipInt;
Procedure SkipHex;

Procedure CheckEOL;
Procedure CheckPartLine (s: string);
Procedure CheckLine (s: string);
Procedure FindLine (s: string);
Procedure FindPartLine (s: string);

IMPLEMENTATION

(********************************* INPUT **********************************)

Var InpFile: text;
    InpBuff: array [1..32*1024] of byte;
    InpCnt:  integer;  { line number }

{---------------------------------------------------------------------------}

Procedure Error (msg: string);
Begin
     raise Exception.Create (msg);
End;

{---------------------------------------------------------------------------}

Procedure GetChar;
Begin
     inc (InpPos);
     if InpPos <= length (InpLine) then InpChar := InpLine [InpPos]
                                   else InpChar := cr;
End;

Function PeekChar (i: integer): char;
Begin
     if i > length (InpLine) then result := ' '
                             else result := InpLine [i];
End;

Procedure GetLine;
Begin
     if eof (InpFile) then begin
        InpEof := true;
        InpLine := '';
        InpPos := 0;
        InpChar := cr;
     end
     else begin
        InpEof := false;
        ReadLn (InpFile, InpLine);
        inc (InpCnt);
        InpPos := 0;
        GetChar;
     end;
End;

Procedure OpenInp (FileName: string);
Begin
     Assign (InpFile, FileName);
     SetTextBuf (InpFile, InpBuff);
     Reset (InpFile);

     InpLine := '';
     InpCnt := 0;
     InpPos := 0;
     InpChar := ' ';
     InpEof := false;

     GetLine;
End;

Procedure CloseInp;
Begin
     Close (InpFile);
End;

Procedure EraseInp;
Begin
     Erase (InpFile);
End;

{--------------------------------------------------------------------------}

Procedure Space;
Begin
     while (InpChar <= ' ') and (InpChar <> cr) do GetChar;
End;

Function GetName: string;
Var  s: string [255];
Begin
     Space;

     s := '';
     while InpChar > ' ' do begin
        s := s + InpChar;
        GetChar;
     end;

     result := s;
End;

Function GetCommaName: string;
Const CommaSep = [',', '.', chr (255)];
Var   s: string [255];
Begin
     Space;

     s := '';
     while InpChar > ' ' do begin
        if not (InpChar in CommaSep) then s := s + InpChar;
        GetChar;
     end;

     result := s;
End;

Function GetInt: integer;
Var  s: string [16];
Begin
     Space;

     s := '';
     while InpChar > ' ' do begin
        s := s + InpChar;
        GetChar;
     end;

     result := StrToInt (s);
End;

Function GetHex: integer;
Var  s: string [16];
Begin
     Space;

     s := '$';
     while InpChar > ' ' do begin
        s := s + InpChar;
        GetChar;
     end;

     result := StrToInt (s);
End;

Function GetAttr: integer;
Begin
     Space;

     result := 0;
     while InpChar > ' ' do begin
        case UpCase (InpChar) of
           'R': inc (result, faReadOnly);
           'A': inc (result, faArchive);
           'H': inc (result, faHidden);
           'S': inc (result, faSysFile);
           'D': inc (result, faDirectory);
           '-', '_', 'W': { nothing };
           else Error ('Unknown attribute ' + InpChar);
        end;
        GetChar;
     end;
End;

{--------------------------------------------------------------------------}

Function GetStdDateTime: integer;
Var  SaveFmt: string;
     SaveDate, SaveTime: char;
     s: string [40];
Begin
     SaveFmt := ShortDateFormat;
     SaveDate := DateSeparator;
     SaveTime := TimeSeparator;

     try
        ShortDateFormat := 'm/d/y';
        DateSeparator := '-';
        TimeSeparator := ':';

        Space;
        s := '';
        while InpChar > ' ' do begin
           s := s + InpChar;
           GetChar;
        end;

        Space;
        s := s + ' ';
        while InpChar > ' ' do begin
           s := s + InpChar;
           GetChar;
        end;

        result := DateTimeToFileDate (StrToDateTime (s));

     finally
        DateSeparator := SaveDate;
        TimeSeparator := SaveTime;
        ShortDateFormat := SaveFmt;
     end;
End;

Function GetFlexDateTime: integer;
Const digit = ['0'..'9'];
Var   SaveFmt: string;
      SaveDate, SaveTime: char;
      s: string [40];
Begin
     SaveFmt := ShortDateFormat;
     SaveDate := DateSeparator;
     SaveTime := TimeSeparator;

     try
        DateSeparator := ' ';
        TimeSeparator := ' ';

        Space;
        s := '';
        while InpChar > ' ' do begin
           if not (InpChar in digit) and (DateSeparator = ' ') then
              DateSeparator := InpChar;
           s := s + InpChar;
           GetChar;
        end;

        Space;
        s := s + ' ';
        while InpChar > ' ' do begin
           if not (InpChar in digit) and (TimeSeparator = ' ') then
              TimeSeparator := InpChar;
           s := s + InpChar;
           GetChar;
        end;

        if InpChar in ['a', 'p'] then begin
           s := s + InpChar;
           GetChar;
        end;

        if DateSeparator = '/' then ShortDateFormat := 'm/d/y'
                               else ShortDateFormat := 'd/m/y';

        try
           result := DateTimeToFileDate (StrToDateTime (s));
        except
           result := 0;
        end;

     finally
        DateSeparator := SaveDate;
        TimeSeparator := SaveTime;
        ShortDateFormat := SaveFmt;
     end;
End;

{--------------------------------------------------------------------------}

Procedure SkipName;
Begin
     GetName;
End;

Procedure SkipInt;
Begin
     GetInt;
End;

Procedure SkipHex;
Begin
     GetHex;
End;

{--------------------------------------------------------------------------}

Procedure CheckEOL;
Begin
     Space;
     if InpChar <> cr then Error ('End of line expected');
End;

Procedure CheckPartLine (s: string);
Begin
     if copy (InpLine, 1, length (s)) <> s then
        Error ('Unexpected line');
     GetLine;
End;

Procedure CheckLine (s: string);
Begin
     if InpLine <> s then
        Error ('Unexpected line');
     GetLine;
End;

Procedure FindLine (s: string);
Begin
     while not InpEof and (InpLine <> s) do
        GetLine;

     if InpEof then Error ('Cannot find first line')
               else GetLine;
End;

Procedure FindPartLine (s: string);
Begin
     while not InpEof and (copy (InpLine, 1, length (s)) <> s) do
        GetLine;

     if InpEof then Error ('Cannot find first line')
               else GetLine;
End;

END.
